//   Copyright (c)  2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/NumTheoryQQ.H"
#include "CoCoA/error.H"

#include <iostream>


namespace CoCoA
{

  ContFracIter::ContFracIter(const BigRat& Q)
  {
    myQuot = num(Q)/den(Q); // floor division
    if (!IsOneDen(Q))
      myFrac = 1/(Q - myQuot); // non-negative!
  }

  const BigInt& ContFracIter::operator*() const
  {
    if (IsEnded(*this)) CoCoA_ERROR(ERR::IterEnded, "ContFracIter::operator*");
    return myQuot;
  }

  ContFracIter& ContFracIter::operator++()
  {
    if (IsEnded(*this)) CoCoA_ERROR(ERR::IterEnded, "ContFracIter::operator++");
    myQuot = num(myFrac)/den(myFrac); // floor division
    if (IsOneDen(myFrac))
      myFrac = 0;
    else
      myFrac = 1/(myFrac - myQuot); // strictly positive!
    return *this;
  }

  ContFracIter ContFracIter::operator++(int)
  {
    ContFracIter prev = *this;
    operator++();
    return prev;
  }


  CFApproximantsIter::CFApproximantsIter(const BigRat& Q):
      myCFIter(Q),
      myApproximant(*myCFIter,1),
      myPrevNum(1),
      myPrevDen(0)
  {}

  CFApproximantsIter::CFApproximantsIter(const ContFracIter& CFIter):
      myCFIter(CFIter),
      myApproximant(*myCFIter,1),
      myPrevNum(1),
      myPrevDen(0)
  {}


  CFApproximantsIter& CFApproximantsIter::operator++()
  {
    if (IsEnded(*this)) CoCoA_ERROR(ERR::IterEnded, "CFApproximantsIter::operator++");
    ++myCFIter;
    // SLUG SLUG SLUG  The next 5 lines are hideously wasteful!!!
    const BigInt NextNum = *myCFIter * num(myApproximant) + myPrevNum;
    const BigInt NextDen = *myCFIter * den(myApproximant) + myPrevDen;
    myPrevNum = num(myApproximant);
    myPrevDen = den(myApproximant);
    myApproximant = BigRat(NextNum, NextDen, BigRat::AlreadyReduced);
//     mpz_mul();
//     mpz_add();
//     mpz_swap();
//     mpz_mul();
//     mpz_add();
//     mpz_swap();
    return *this;
  }

  CFApproximantsIter CFApproximantsIter::operator++(int)
  {
    CFApproximantsIter prev = *this;
    operator++();
    return prev;
  }


  std::ostream& operator<<(std::ostream& out, const ContFracIter& CFIter)
  {
    out << "ContFracIter(...)";
    return out;
  }

  std::ostream& operator<<(std::ostream& out, const CFApproximantsIter& CFAIter)
  {
    out << "CFApproximantsIter(...)";
    return out;
  }


  BigRat CFApprox(const BigRat& q, const BigRat& MaxRelErr)
  {
    // Simple rather than superfast.
    if (MaxRelErr < 0 || MaxRelErr > 1) CoCoA_ERROR(ERR::BadArg, "CFApprox: relative error must be between 0 and 1");
    if (IsZero(q) || IsZero(MaxRelErr)) return q;
    if (MaxRelErr == 1) return BigRat(0);
    const BigRat MaxAbsErr = abs(q*MaxRelErr);
    if (MaxAbsErr >= 1) return BigRat(floor(q),1);
    CFApproximantsIter CFAIter(q);
    while (abs(q - *CFAIter) > MaxAbsErr)
      ++CFAIter;
    return *CFAIter;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/NumTheoryQQ.C,v 1.7 2011/08/24 10:28:49 bigatti Exp $
// $Log: NumTheoryQQ.C,v $
// Revision 1.7  2011/08/24 10:28:49  bigatti
// -- renamed QQ --> BigRat
// -- sorted #include
//
// Revision 1.6  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.5  2011/06/23 16:05:12  abbott
// Removed single arg ctor QQ(MachineInteger).
//
// Revision 1.4  2011/02/25 12:05:40  abbott
// Added new fn CFApprox (as in CoCoA4)
//
// Revision 1.3  2010/05/07 14:57:52  abbott
// Two main changes:
//   power(QQ,ZZ) now allows negative exponent
//   renamed QQ::AlreadyNormalized to QQ::AlreadyReduced
//           (and allowed denoms to be negative; the ctor then makes them positive).
//
// Revision 1.2  2010/03/22 18:09:33  abbott
// Corrected copyright year.
//
// Revision 1.1  2010/03/18 16:35:08  abbott
// Added NumTheoryQQ files for number theory operations involving QQs.
//
