//   Copyright (c)  2010  John Abbott

//   This file is part of the source of CoCoALib, the CoCoA Library.

//   CoCoALib is free software: you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, either version 3 of the License, or
//   (at your option) any later version.

//   CoCoALib is distributed in the hope that it will be useful,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//   GNU General Public License for more details.

//   You should have received a copy of the GNU General Public License
//   along with CoCoALib.  If not, see <http://www.gnu.org/licenses/>.


#include "CoCoA/RandomSource.H"
#include "CoCoA/error.H"
#include "CoCoA/utils.H"

#include <iostream>
//using std::ostream;
#include <limits>
using std::numeric_limits;


namespace CoCoA
{

  RandomSource::RandomSource(const MachineInteger& seed):
      mySeed(seed)
  {
    gmp_randinit_mt(myState);
    gmp_randseed(myState, mpzref(mySeed));
  }


  RandomSource::RandomSource(const BigInt& seed):
      mySeed(seed)
  {
    gmp_randinit_mt(myState);
    gmp_randseed(myState, mpzref(mySeed));
  }


  RandomSource::RandomSource(const RandomSource& copy):
      mySeed(copy.mySeed)
  {
    gmp_randinit_set(myState, copy.myState);
  }


  RandomSource& RandomSource::operator=(const RandomSource& rhs)
  {
    gmp_randclear(myState);                 // Together these effect an assignment
    gmp_randinit_set(myState, rhs.myState); //
    mySeed = rhs.mySeed;
    return *this;
  }


  RandomSource::~RandomSource()
  {
    gmp_randclear(myState);
  }


  long RandomLong(RandomSource& src)
  {
    const unsigned long rnd = gmp_urandomb_ui(src.myState, numeric_limits<unsigned long>::digits);
    return ULong2Long(rnd);
  }


  long RandomLong(RandomSource& src, const MachineInteger& lwb, const MachineInteger& upb)
  {
    if (!IsSignedLong(lwb) || !IsSignedLong(upb))
      CoCoA_ERROR(ERR::BadArg, "RandomLong: lwb & upb must each fit into a long");
    const long lo = AsSignedLong(lwb);
    const long hi = AsSignedLong(upb);
    if (lo > hi) CoCoA_ERROR(ERR::BadArg, "RandomLong: must have lwb <= upb");
    const unsigned long range = (1ul + hi) - lo; // all arith is between ulongs.
    if (range == 0) return RandomLong(src);
    return ULong2Long(lo + gmp_urandomm_ui(src.myState, range)); // lo is silently converted to ulong
  }


  BigInt RandomBigInt(RandomSource& src, const MachineInteger& lwb, const MachineInteger& upb)
  {
    return RandomBigInt(src, BigInt(lwb), BigInt(upb));
  }

  BigInt RandomBigInt(RandomSource& src, const MachineInteger& lwb, const BigInt& upb)
  {
    return RandomBigInt(src, BigInt(lwb), upb);
  }

  BigInt RandomBigInt(RandomSource& src, const BigInt& lwb, const MachineInteger& upb)
  {
    return RandomBigInt(src, lwb, BigInt(upb));
  }

  BigInt RandomBigInt(RandomSource& src, const BigInt& lwb, const BigInt& upb)
  {
    if (lwb > upb) CoCoA_ERROR(ERR::BadArg, "RandomBigInt: must have lwb <= upb");
    BigInt sample;
    mpz_urandomm(mpzref(sample), src.myState, mpzref(upb-lwb+1));
    return lwb + sample;
  }

// NOTE (2011-04-26): this is significantly faster than the fn above.
//   BigInt RandomBigIntmod(RandomSource& src, const BigInt& M)
//   {
//     if (M <= 0) CoCoA_ERROR(ERR::BadArg, "RandomBigIntmod: must have M > 0");
//     BigInt sample;
//     mpz_urandomm(mpzref(sample), src.myState, mpzref(M));
//     return sample;
//   }


  void reseed(RandomSource& src, const MachineInteger& seed)
  {
    reseed(src, BigInt(seed));
  }

  void reseed(RandomSource& src, const BigInt& seed)
  {
    src.mySeed = seed;
    gmp_randseed(src.myState, mpzref(src.mySeed));
    // zero any counters
  }


  std::ostream& operator<<(std::ostream& out, const RandomSource& src)
  {
    out << "RandomSource(seed=" << src.mySeed << ", state=unprintable)";
    return out;
  }

} // end of namespace CoCoA


// RCS header/log in the next few lines
// $Header: /Volumes/Home_1/cocoa/cvs-repository/CoCoALib-0.99/src/AlgebraicCore/RandomSource.C,v 1.4 2011/08/14 15:52:17 abbott Exp $
// $Log: RandomSource.C,v $
// Revision 1.4  2011/08/14 15:52:17  abbott
// Changed ZZ into BigInt (phase 1: just the library sources).
//
// Revision 1.3  2011/05/03 09:52:50  abbott
// Changed default initial seed to 0 (was previsouly 1).
// Added reseeding functions.
// Seeding from a ZZ is now possible.
//
// Revision 1.2  2011/03/14 10:26:21  abbott
// Changed size_t into long.
//
// Revision 1.1  2010/12/17 16:03:15  abbott
// Added new class RandomSource.
//
//
